unit Containers;

{$mode objfpc}{$H+}
{$modeswitch advancedrecords}

interface

uses
  Classes, SysUtils;

type

  { TDynArrayData - Contains the actual data for a dynamic array as a managed
    record to allow for automatic finalization }
  generic TDynArrayData<T> = record
  type
    // Pointer to the data type
    PT = ^T;
  private
    // The items in the array
    Items: Pointer;
    // Number of items
    Count: SizeInt;
    // Allocated capacity
    Capacity: SizeInt;
  private
    // Clear the data
    procedure Clear;
    // Set the capacity and reallocate memory if needed
    procedure SetCapacity(ACapacity: SizeInt);
    // Set the count and update capacity if needed
    procedure SetCount(ACount: SizeInt);
  public
    // Management operators
    class operator Initialize(var DAD: specialize TDynArrayData<T>);
    class operator Finalize(var DAD: specialize TDynArrayData<T>);
  end;

  { TDynArrayView - A read-only dynamic array that can be used to expose a
    TDynArray as an immutable object }
  generic TDynArrayView<T> = object
  type
    // Pointer to the data type
    PT = ^T;
    // Callback function for sorting
    TSorter = function(const A, B: T; Data: Pointer): SizeInt;
  private
    // Property storage
    Data: specialize TDynArrayData<T>;
    function GetFirst: T;
    function GetFirstPointer: PT;
    function GetItems(AIndex: SizeInt): T;
    function GetItemPointers(AIndex: SizeInt): PT;
    function GetLast: T;
    function GetLastPointer: PT;
  public
    // Returns the index of the given item's first match or -1 if not found
    function IndexOf(Item: T): SizeInt;
    // Returns the index of the given item's last match or -1 if not found
    function LastIndexOf(Item: T): SizeInt;
    // Returns true if the container contains the given item
    function Contains(Item: T): Boolean;
    // Returns true if the container is empty
    function IsEmpty: Boolean;
    // Returns true if the container has items
    function HasItems: Boolean;
  public
    // Get the items in the container
    property Items[AIndex: SizeInt]: T read GetItems; default;
    // Get pointers to the items in the container
    property ItemPointers[AIndex: SizeInt]: PT read GetItemPointers;
    // Get the number of items in the container
    property Count: SizeInt read Data.Count;
    // Get the amount of allocated memory in terms of items
    property Capacity: SizeInt read Data.Capacity;
    // Get the first item in the container
    property First: T read GetFirst;
    // Get the last item in the container
    property Last: T read GetLast;
    // Get a pointer to the first item in the container
    property FirstPointer: PT read GetFirstPointer;
    // Get a pointer to the last item in the container
    property LastPointer: PT read GetLastPointer;
  end;

  { TDynArray - Provides a dynamic array that can also act as a stack and
    decouples number of items from allocated memory to minimize allocations }
  generic TDynArray<T> = object(specialize TDynArrayView<T>)
  private
    procedure SetCapacity(ACapacity: SizeInt);
    procedure SetCount(ACount: SizeInt);
    procedure SetItems(AIndex: SizeInt; AValue: T);
    procedure SetFirst(const AValue: T);
    procedure SetLast(const AValue: T);
    function GetRawData: PT;
  private
    // Used by Sort
    function Partition(L, H: SizeInt; ASorter: TSorter; AData: Pointer): SizeInt;
    procedure QuickSort(L, H: SizeInt; ASorter: TSorter; AData: Pointer);
  public
    // Ensure the array has *at least* the given capacity
    procedure EnsureCapacity(ACapacity: SizeInt);
    // Assign this array with the items of the given array
    function Assign(constref Other: specialize TDynArrayView<T>): SizeInt;
    // Assign this array with the items of the given array using a raw memory
    // copy that can be faster than normal Assign but wont call any operators
    procedure AssignRaw(constref Other: specialize TDynArrayView<T>);
    // Add an item at the end of the the array, returns the new item's index
    function Add(Item: T): SizeInt;
    // Add a new item at the end of the array and return a pointer to it
    function AddNew: PT;
    // Add all the items from the given array to this array and return the index
    // of the first item added
    function AddAll(constref Other: specialize TDynArrayView<T>): SizeInt;
    // Add all the items from the given array to this array and return the index
    // of the first item added
    function AddAll(constref Other: array of T): SizeInt;
    // Add the given item if there isn't already any with the same value,
    // returns the new or existing item's index
    function AddUnique(Item: T): SizeInt;
    // Insert an item at the given position of the array (Count is same as Add)
    procedure Insert(Index: SizeInt; Item: T);
    // Removes the item at the given index
    procedure RemoveAt(Index: SizeInt);
    // Remove the given item if it exists in the container
    procedure Remove(Item: T);
    // Removes all instances of the given item from the container
    procedure RemoveAll(Item: T);
    // Removes the item at the given index by swapping it with the last item and resizing
    procedure RemoveViaSwapAt(Index: SizeInt);
    // Removes the given item by swapping it with the last item and resizing
    procedure RemoveViaSwap(Item: T);
    // Replace all instances of the old item with the new item
    procedure Replace(Item, Replacement: T);
    // Replace the first instance of the old item with the new item
    procedure ReplaceFirst(Item, Replacement: T);
    // Replace all last instance of the old item with the new item
    procedure ReplaceLast(Item, Replacement: T);
    // Removes all items and releases any allocated memory
    procedure Clear;
    // Removes all items without releasing any memory
    procedure Reset;
    // Swaps the items at the given indices
    procedure Swap(Index1, Index2: SizeInt);
    // Removes the last item and returns it
    function Pop: T;
    // Removes the last item
    procedure PopAndIgnore;
    // Resize the capacity to be the same as the count of items in the container
    procedure Pack;
    // Sort the array using the given sorting function
    procedure Sort(ASorter: TSorter; AData: Pointer);
    // Quickly swaps all the items between this array and the given array by
    // swapping the pointers and other data instead of individual items
    procedure SwapItemsWith(var Other: specialize TDynArray<T>);
  public
    // Set/Get the items in the container
    property Items[AIndex: SizeInt]: T read GetItems write SetItems; default;
    // Set/Get the number of items in the container
    property Count: SizeInt read Data.Count write SetCount;
    // Set/Get the amount of allocated memory in terms of items
    property Capacity: SizeInt read Data.Capacity write SetCapacity;
    // Get/Set the first item in the container
    property First: T read GetFirst write SetFirst;
    // Get/Set the last item in the container
    property Last: T read GetLast write SetLast;
    // Get the items as raw data
    property RawData: PT read GetRawData;
    // Equality operator
    //class operator=(constref A, B: specialize TDynArray<T>): Boolean;
  end;

  { A dynamic array of raw pointers }
  TPointerDynArray = specialize TDynArray<Pointer>;
  { A dynamic array view of raw pointers }
  TPointerDynArrayView = ^specialize TDynArrayView<Pointer>;
  { A dynamic array of strings }
  TStringDynArray = specialize TDynArray<string>;
  { A dynamic array view of strings }
  TStringDynArrayView = ^specialize TDynArrayView<string>;
  { A dynamic array of integers }
  TIntegerDynArray = specialize TDynArray<Integer>;
  { A dynamic array view of integers }
  TIntegerDynArrayView = ^specialize TDynArrayView<Integer>;
  { A dynamic array of booleans }
  TBooleanDynArray = specialize TDynArray<Boolean>;
  { A dynamic array view of booleans }
  TBooleanDynArrayView = ^specialize TDynArrayView<Boolean>;
  { A dynamic array of bytes }
  TByteDynArray = specialize TDynArray<Byte>;
  { A dynamic array view of bytes }
  TByteDynArrayView = ^specialize TDynArrayView<Byte>;
  { A dynamic array of characters }
  TCharDynArray = specialize TDynArray<Char>;
  { A dynamic array view of characters }
  TCharDynArrayView = ^specialize TDynArrayView<Char>;
  { A dynamic array of methods }
  TMethodDynArray = specialize TDynArray<TMethod>;
  { A dynamic array view of methods }
  TMethodDynArrayView = ^specialize TDynArrayView<TMethod>;

  { TPair - A generic container for a pair of values }
  generic TPair<F,S> = record
    // First value of the pair
    First: F;
    // Second value of the pair
    Second: S;
    // Make a pair
    class function Make(const AFirst: F; const ASecond: S): specialize TPair<F,S>; static;
    // Equality operator
    class operator=(constref A, B: specialize TPair<F,S>): Boolean;
  end;

  { A pair of bytes }
  TBytePair = specialize TPair<Byte,Byte>;
  { A pair of strings }
  TStringPair = specialize TPair<string,string>;
  { A pair of pointers }
  TPointerPair = specialize TPair<Pointer,Pointer>;
  { A string and pointer pair }
  TStringPointerPair = specialize TPair<string,Pointer>;

  { TTriplet - A generic container for a value triplet }
  generic TTriplet<F,S,T> = record
    // First value of the triplet
    First: F;
    // Second value of the triplet
    Second: S;
    // Third value of the triplet
    Third: T;
    // Make a triplet
    class function Make(const AFirst: F; const ASecond: S; const AThird: T): specialize TTriplet<F,S,T>; static;
    // Equality operator
    class operator=(constref A, B: specialize TTriplet<F,S,T>): Boolean;
  end;

  { TDictionary - A dictionary that maps K values to V values }
  generic TDictionary<K,V> = record
  type
    // A pair in the dictionary
    TKVPair = specialize TPair<K,V>;
    // Pointer to the value data type
    PV = ^V;
  private
    // Property storage
    FItems: specialize TDynArray<TKVPair>;
    function GetCount: SizeInt;
    function GetKeys(AIndex: SizeInt): K;
    function GetValues(AIndex: SizeInt): V;
    function GetPairs(AIndex: SizeInt): TKVPair;
    function GetItems(const KVal: K): V;
    procedure SetItems(const KVal: K; const VVal: V);
  public
    // Clear the dictionary
    procedure Clear;
    // Add a new value or replace an existing value in the dictionary
    procedure Add(const KVal: K; const VVal: V);
    // Remove an existing value from the dictionary
    procedure Remove(const KVal: K);
    // Return the pointer to the value for the given key or nil if not found
    function Find(const KVal: K): PV;
  public
    // Number of items in the dictionary
    property Count: SizeInt read GetCount;
    // The keys in the dictionary
    property Keys[AIndex: SizeInt]: K read GetKeys;
    // The values in the dictionary
    property Values[AIndex: SizeInt]: V read GetValues;
    // The key-value pairs in the dictionary
    property Pairs[AIndex: SizeInt]: TKVPair read GetPairs;
    // The values accessible via brackets (will insert new values if needed)
    property Items[KVal: K]: V read GetItems write SetItems; default;
  public
    // Equality operator
    class operator=(constref A, B: specialize TDictionary<K,V>): Boolean;
  end;

// Convert a Free Pascal TStringArray to a TStringDynArray
function StringArrayToStringDynArray(const SA: TStringArray): TStringDynArray;
// Compare two dynamic arrays for equality
generic function CompareDynArray<T>(const A, B: specialize TDynArray<T>): Boolean;

// Equality operators for some FPC-provided types
operator=(const A, B: TMethod): Boolean;

implementation


{ Functions }
function StringArrayToStringDynArray(const SA: TStringArray): TStringDynArray;
var
  I: Integer;
begin
  Result.Count:=Length(SA);
  for I:=0 to High(SA) do Result[I]:=SA[I];
end;

generic function CompareDynArray<T>(const A, B: specialize TDynArray<T>): Boolean;
var
  I: SizeInt;
begin
  Result:=A.Count=B.Count;
  if not Result then Exit;
  for I:=0 to A.Count - 1 do
    if A[I] <> B[I] then Exit(False);
end;

operator=(const A, B: TMethod): Boolean;
begin
  Result:=(A.Code=B.Code) and (A.Data=B.Data);
end;

{ TDynArrayData }
procedure TDynArrayData.Clear;
begin
  Count:=0;
  SetCapacity(0);
end;

procedure TDynArrayData.SetCapacity(ACapacity: SizeInt);
begin
  if Capacity=ACapacity then Exit;
  if ACapacity < Count then ACapacity:=Count;
  // If the capacity is shrinking, finalize the items to be released
  if ACapacity < Capacity then
    Finalize(PT(Items)[ACapacity], Capacity - ACapacity);
  ReAllocMem(Items, ACapacity*SizeOf(T));
  // If the capacity is growing, initialize the new items
  if ACapacity > Capacity then
    Initialize(PT(Items)[Capacity], ACapacity - Capacity);
  Capacity:=ACapacity;
end;

procedure TDynArrayData.SetCount(ACount: SizeInt);
begin
  if Count=ACount then Exit;
  Count:=ACount;
  if ACount > Capacity then
    if Capacity=0 then
      SetCapacity(Count)
    else if Capacity*SizeOf(T) < 16*1024*1024 then begin
      if Count > Capacity*2 then
        SetCapacity(Count)
      else
        SetCapacity(Capacity*2)
    end else if SizeOf(T) < 64 then
      {%H-}SetCapacity(Count + 256)
    else
      {%H-}SetCapacity(Count + 16);
end;

class operator TDynArrayData.Initialize(var DAD: specialize TDynArrayData<T>);
begin
  DAD:=Default(specialize TDynArrayData<T>);
end;

class operator TDynArrayData.Finalize(var DAD: specialize TDynArrayData<T>);
begin
  DAD.Clear;
end;

{ TDynArrayView }
function TDynArrayView.GetFirst: T;
begin
  Assert(Count > 0, 'Tried to access the first item in an empty dynamic array');
  Result:=PT(Data.Items)[0];
end;

function TDynArrayView.GetFirstPointer: PT;
begin
  Assert(Count > 0, 'Tried to access the first item in an empty dynamic array');
  Result:=@PT(Data.Items)[0];
end;

function TDynArrayView.GetItems(AIndex: SizeInt): T;
begin
  Assert((AIndex >= 0) and (AIndex < Count), 'Invalid dynamic array item index');
  Result:=PT(Data.Items)[AIndex];
end;

function TDynArrayView.GetItemPointers(AIndex: SizeInt): PT;
begin
  Assert((AIndex >= 0) and (AIndex < Count), 'Invalid dynamic array item index');
  Result:=@PT(Data.Items)[AIndex];
end;

function TDynArrayView.GetLast: T;
begin
  Assert(Count > 0, 'Tried to access the last item in an empty dynamic array');
  Result:=Items[Count - 1];
end;

function TDynArrayView.GetLastPointer: PT;
begin
  Assert(Count > 0, 'Tried to access the last item in an empty dynamic array');
  Result:=@PT(Data.Items)[Count - 1];
end;

function TDynArrayView.IndexOf(Item: T): SizeInt;
var
  I: SizeInt;
begin
  for I:=0 to Count - 1 do
    if Item=Items[I] then
      Exit(I);
  Result:=-1;
end;

function TDynArrayView.LastIndexOf(Item: T): SizeInt;
var
  I: SizeInt;
begin
  for I:=Count - 1 downto 0 do
    if Item=Items[I] then
      Exit(I);
  Result:=-1;
end;

function TDynArrayView.Contains(Item: T): Boolean;
begin
  Result:=IndexOf(Item) <> -1;
end;

function TDynArrayView.IsEmpty: Boolean;
begin
  Result:=Data.Count=0;
end;

function TDynArrayView.HasItems: Boolean;
begin
  Result:=Data.Count > 0;
end;

{ TDynArray }
procedure TDynArray.SetCapacity(ACapacity: SizeInt);
begin
  Data.SetCapacity(ACapacity);
end;

procedure TDynArray.SetCount(ACount: SizeInt);
begin
  Data.SetCount(ACount);
end;

procedure TDynArray.SetItems(AIndex: SizeInt; AValue: T);
begin
  Assert((AIndex >= 0) and (AIndex < Count), 'Invalid dynamic array item index');
  PT(Data.Items)[AIndex]:=AValue;
end;

function TDynArray.Partition(L, H: SizeInt; ASorter: TSorter; AData: Pointer): SizeInt;
var
  Pivot: T;
  J: SizeInt;
begin
  Pivot:=GetItems(H);
  Result:=L;
  for J:=L to H - 1 do
    if ASorter(GetItems(J), Pivot, AData) < 0 then begin
      if Result <> J then Swap(Result, J);
      Inc(Result);
    end;
  Swap(Result, H);
end;

procedure TDynArray.QuickSort(L, H: SizeInt; ASorter: TSorter; AData: Pointer);
var
  P: SizeInt;
begin
  if L < H then begin
    P:=Partition(L, H, ASorter, AData);
    QuickSort(L, P - 1, ASorter, AData);
    QuickSort(P + 1, H, ASorter, AData);
  end;
end;

procedure TDynArray.SetFirst(const AValue: T);
begin
  Assert(Count > 0, 'Tried to set the first item of an empty dynamic array');
  PT(Data.Items)[0]:=AValue;
end;

procedure TDynArray.SetLast(const AValue: T);
begin
  Assert(Count > 0, 'Tried to set the last item of an empty dynamic array');
  PT(Data.Items)[Data.Count - 1]:=AValue;
end;

function TDynArray.GetRawData: PT;
begin
  Result:=PT(Data.Items);
end;

procedure TDynArray.EnsureCapacity(ACapacity: SizeInt);
begin
  if Data.Capacity < ACapacity then SetCapacity(ACapacity);
end;

function TDynArray.Assign(constref Other: specialize TDynArrayView<T>): SizeInt;
begin
  Reset;
  Result:=AddAll(Other);
end;

procedure TDynArray.AssignRaw(constref Other: specialize TDynArrayView<T>);
begin
  Count:=Other.Count;
  if Count > 0 then Move(Other.Data.Items^, Data.Items^, Count*SizeOf(T));
end;

function TDynArray.Add(Item: T): SizeInt;
begin
  Result:=Count;
  Insert(Count, Item);
end;

function TDynArray.AddNew: PT;
begin
  SetCount(Count + 1);
  Result:=LastPointer;
end;

function TDynArray.AddAll(constref Other: specialize TDynArrayView<T>): SizeInt;
var
  I: SizeInt;
begin
  Result:=Count;
  if Other.HasItems then begin
    EnsureCapacity(Count + Other.Count);
    for I:=0 to Other.Count - 1 do
      Add(Other[I]);
  end;
end;

function TDynArray.AddAll(constref Other: array of T): SizeInt;
var
  I: SizeInt;
begin
  Result:=Count;
  if Length(Other) > 0 then begin
    EnsureCapacity(Count + Length(Other));
    for I:=0 to High(Other) do
      Add(Other[I]);
  end;
end;

function TDynArray.AddUnique(Item: T): SizeInt;
begin
  Result:=IndexOf(Item);
  if Result=-1 then Result:=Add(Item);
end;

procedure TDynArray.Insert(Index: SizeInt; Item: T);
var
  I: SizeInt;
begin
  Assert((Index >= 0) and (Index <= Count), 'Invalid dynamic array insertion index');
  Count:=Count + 1;
  for I:=Data.Count - 1 downto Index + 1 do
    PT(Data.Items)[I]:=PT(Data.Items)[I - 1];
  PT(Data.Items)[Index]:=Item;
end;

procedure TDynArray.RemoveAt(Index: SizeInt);
var
  I: SizeInt;
begin
  Assert((Index >= 0) and (Index < Count), 'Invalid dynamic array item index');
  for I:=Index to Count - 2 do Items[I]:=Items[I + 1];
  Count:=Count - 1;
end;

procedure TDynArray.Remove(Item: T);
var
  Index: SizeInt;
begin
  Index:=IndexOf(Item);
  if Index <> -1 then RemoveAt(Index);
end;

procedure TDynArray.RemoveAll(Item: T);
var
  I: SizeInt;
begin
  for I:=Count - 1 downto 0 do
    if Items[I]=Item then RemoveAt(I);
end;

procedure TDynArray.RemoveViaSwapAt(Index: SizeInt);
begin
  Assert((Index >= 0) and (Index < Count), 'Invalid dynamic array item index');
  if Index < Count - 1 then Swap(Index, Count - 1);
  Dec(Data.Count);
end;

procedure TDynArray.RemoveViaSwap(Item: T);
var
  Index: SizeInt;
begin
  Index:=IndexOf(Item);
  if Index <> -1 then RemoveViaSwapAt(Index);
end;

procedure TDynArray.Replace(Item, Replacement: T);
var
  I: SizeInt;
begin
  for I:=0 to Count - 1 do
    if Items[I]=Item then Items[I]:=Replacement;
end;

procedure TDynArray.ReplaceFirst(Item, Replacement: T);
var
  I: SizeInt;
begin
  for I:=0 to Count - 1 do
    if Items[I]=Item then begin
      Items[I]:=Replacement;
      Break;
    end;
end;

procedure TDynArray.ReplaceLast(Item, Replacement: T);
var
  I: SizeInt;
begin
  for I:=Count - 1 downto 0 do
    if Items[I]=Item then begin
      Items[I]:=Replacement;
      Break;
    end;
end;

procedure TDynArray.Clear;
begin
  Data.Clear;
end;

procedure TDynArray.Reset;
begin
  Count:=0;
end;

procedure TDynArray.Swap(Index1, Index2: SizeInt);
var
  Tmp: T;
begin
  Assert((Index1 >= 0) and (Index1 < Count), 'Invalid first index');
  Assert((Index2 >= 0) and (Index2 < Count), 'Invalid second index');
  Tmp:=PT(Data.Items)[Index1];
  PT(Data.Items)[Index1]:=PT(Data.Items)[Index2];
  PT(Data.Items)[Index2]:=Tmp;
end;

function TDynArray.Pop: T;
begin
  Assert(Count > 0, 'Tried to pop from an empty array');
  Result:=PT(Data.Items)[Count - 1];
  Dec(Data.Count);
end;

procedure TDynArray.PopAndIgnore;
begin
  Assert(Count > 0, 'Tried to pop from an empty array');
  Dec(Data.Count);
end;

procedure TDynArray.Pack;
begin
  Capacity:=Count;
end;

procedure TDynArray.Sort(ASorter: TSorter; AData: Pointer);
begin
  if Count > 1 then QuickSort(0, Count - 1, ASorter, AData);
end;

procedure TDynArray.SwapItemsWith(var Other: specialize TDynArray<T>);
var
  TmpItems: Pointer;
  TmpCount: SizeInt;
  TmpCapacity: SizeInt;
begin
  if @Other <> @Self then begin
    TmpItems:=Data.Items;
    TmpCount:=Data.Count;
    TmpCapacity:=Data.Capacity;
    Data.Items:=Other.Data.Items;
    Data.Count:=Other.Data.Count;
    Data.Capacity:=Other.Data.Capacity;
    Other.Data.Items:=TmpItems;
    Other.Data.Count:=TmpCount;
    Other.Data.Capacity:=TmpCapacity;
  end;
end;

//class operator TDynArray.=(constref A, B: specialize TDynArray<T>): Boolean;
//var
//  I: SizeInt;
//begin
//  if A.Count <> B.Count then Exit(False);
//  for I:=0 to A.Count - 1 do
//    if PT(A.FItems)[I] <> PT(B.FItems)[I] then Exit(False);
//  Result:=True;
//end;

{ TPair }
class function TPair.Make(const AFirst: F; const ASecond: S): specialize TPair<F, S>;
begin
  Result.First:=AFirst;
  Result.Second:=ASecond;
end;

class operator TPair.=(constref A, B: specialize TPair<F,S>): Boolean;
begin
  Result:=(A.First=B.First) and (A.Second=B.Second);
end;

{ TTriplet }
class function TTriplet.Make(const AFirst: F; const ASecond: S; const AThird: T): specialize TTriplet<F, S, T>;
begin
  Result.First:=AFirst;
  Result.Second:=ASecond;
  Result.Third:=AThird;
end;

class operator TTriplet.=(constref A, B: specialize TTriplet<F,S,T>): Boolean;
begin
  Result:=(A.First=B.First) and (A.Second=B.Second) and (A.Third=B.Third);
end;

{ TDictionary }
function TDictionary.GetCount: SizeInt;
begin
  Result:=FItems.Count;
end;

function TDictionary.GetKeys(AIndex: SizeInt): K;
begin
  Result:=FItems.ItemPointers[AIndex]^.First;
end;

function TDictionary.GetValues(AIndex: SizeInt): V;
begin
  Result:=FItems.ItemPointers[AIndex]^.Second;
end;

function TDictionary.GetPairs(AIndex: SizeInt): TKVPair;
begin
  Result:=FItems.ItemPointers[AIndex]^;
end;

function TDictionary.GetItems(const KVal: K): V;
var
  Ptr: PV;
begin
  Ptr:=Find(KVal);
  if Assigned(Ptr) then Result:=Ptr^ else Result:=Default(V);
end;

procedure TDictionary.SetItems(const KVal: K; const VVal: V);
var
  Ptr: PV;
  NewPair: ^TKVPair;
begin
  Ptr:=Find(KVal);
  if Assigned(Ptr) then Ptr^:=VVal else begin
    NewPair:=FItems.AddNew;
    NewPair^.First:=KVal;
    Newpair^.Second:=VVal;
  end;
end;

procedure TDictionary.Clear;
begin
  FItems.Clear;
end;

procedure TDictionary.Add(const KVal: K; const VVal: V);
begin
  SetItems(KVal, VVal);
end;

procedure TDictionary.Remove(const KVal: K);
var
  I: SizeInt;
begin
  for I:=0 to FItems.Count - 1 do
    if FItems.ItemPointers[I]^.First=KVal then begin
      FItems.RemoveAt(I);
      Exit;
    end;
end;

function TDictionary.Find(const KVal: K): PV;
var
  I: SizeInt;
begin
  for I:=0 to FItems.Count - 1 do
    if FItems.ItemPointers[I]^.First=KVal then Exit(@FItems.ItemPointers[I]^.Second);
  Result:=nil;
end;


class operator TDictionary.=(constref A, B: specialize TDictionary<K, V>): Boolean;
var
  I: SizeInt;
begin
  if @A=@B then Exit(True);
  if A.FItems.Count <> B.FItems.Count then Exit(False);
  Result:=True;
  for I:=0 to A.FItems.Count do
    if A.FItems.ItemPointers[I]^ <> B.FItems.ItemPointers[I]^ then Exit(False);
end;

end.

