unit Documents;
{$mode objfpc}{$H+}
{$scopedenums on}
{$WARN 5024 off : Parameter "$1" not used}
interface

uses
  Classes, SysUtils, Controls, Containers;

type
  { Forward class declarations }
  TDocView = class;
  TDocument = class;
  TDocumentManager = class;
  TDocViewClass = class of TDocView;

  { Exceptions }
  EDocumentError = class(Exception);

  { Collections }
  TDocumentDynArray = specialize TDynArray<TDocument>;
  TDocViewDynArray = specialize TDynArray<TDocView>;

  { Events }
  TDocumentEvent = procedure(Sender: TObject; ADocument: TDocument) of object;
  TDocumentFileEvent = procedure(Sender: TObject; ADocument: TDocument; AFileName: string) of object;
  TFileNameEvent = procedure(Sender: TObject; AFileName: string) of object;
  TViewEvent = procedure(Sender: TObject; AView: TDocView) of object;
  TDocumentLastViewRemovedEvent = procedure(Sender: TObject; var FreeDocument: Boolean) of object;
  TDocManagerLastViewRemovedEvent = procedure(Sender: TObject; ADocument: TDocument; var FreeDocument: Boolean) of object;

  { TDocManagerListener - Listener for document events in the document manager }
  TDocManagerListener = class
  public
    // A view was added
    procedure ViewAdded(ADocument: TDocument; AView: TDocView); virtual;
    // A view was removed
    procedure ViewRemoved(ADocument: TDocument; AView: TDocView); virtual;
    // The last view was removed, if FreeDocument is True the document will be
    // automatically freed and removed from the document manager
    procedure LastViewRemoved(ADocument: TDocument; var FreeDocument: Boolean); virtual;
    // A document was unregistered from the document manager
    procedure DocumentUnregistered(ADocument: TDocument); virtual;
    // A document was registered with the document manager
    procedure DocumentRegistered(ADocument: TDocument); virtual;
    // A document was modified
    procedure DocumentModified(ADocument: TDocument); virtual;
    // A document's filename was changed
    procedure DocumentFileNameChanged(ADocument: TDocument); virtual;
    // A document was saved to the given filename
    procedure DocumentSaved(ADocument: TDocument; AFileName: string); virtual;
    // A document was loaded from the given filename
    procedure DocumentLoaded(ADocument: TDocument; AFileName: string); virtual;
  end;

  { TDocView - Base class for document views }
  TDocView = class
  private
    // Property storage
    FDocument: TDocument;
  protected
    // Called to initialize the view after it has been created by the document
    // and associatde with it but before any event handlers have been notified
    procedure InitializeView; virtual;
    // Called when the document's modified state is changed
    procedure Modified; virtual;
    // Called when part or whole of the document is invalidated.  For complex
    // documents the "Detail" parameter describes what changed (this depends
    // on the underlying document)
    procedure Invalidated(Detail: TObject); virtual;
    // Called when the document's filename changes
    procedure FileNameChanged; virtual;
    // Called after the document is saved to the given filename
    procedure Saved(AFileName: string); virtual;
    // Called after the document has been loaded from the given filename
    procedure Loaded(AFileName: string); virtual;
  public
    // Create the view
    constructor Create; virtual;
    // Destroy the view
    destructor Destroy; override;
    // Refresh the view (subclasses must make this *do* something)
    procedure Refresh; virtual;
    // Invalidate all document views
    procedure Invalidate(Detail: TObject=nil);
    // Refresh all document views
    procedure RefreshViews;
    // Returns true if this view is the last view remaining for the document
    function IsLastView: Boolean;
  public
    // The document this view represents
    property Document: TDocument read FDocument;
  end;

  { TControlDocView - Base class for document views that provide a control to
    represent the view }
  TControlDocView = class(TDocView)
  protected
    // Return the control for this view
    function GetControl: TControl; virtual; abstract;
  public
    // Returns True if there is a control in this view (the default
    // implemementation calls GetControl to check if it returns Nil but a
    // subclass can override this to avoid creating a control on demand)
    function HasControl: Boolean; virtual;
  public
    // The control that represents the view
    property Control: TControl read GetControl;
  end;

  { TDocument - Base class for documents that can have multiple views }
  TDocument = class
  private type
    TDocFlag = (Modified, Destroying);
    TDocFlags = set of TDocFlag;
  private
    // Property storage
    FFlags: TDocFlags;
    FViews: TDocViewDynArray;
    FFileName: string;
    FDocumentManager: TDocumentManager;
    FOnViewAdd: TViewEvent;
    FOnViewRemove: TViewEvent;
    FOnLastViewRemove: TDocumentLastViewRemovedEvent;
    FOnModify: TNotifyEvent;
    FOnFileNameChange: TNotifyEvent;
    FOnSave: TFileNameEvent;
    FOnLoad: TFileNameEvent;
    function GetModified: Boolean;
    function GetViews(AIndex: Integer): TDocView; inline;
    function GetViewCount: Integer; inline;
    procedure SetFileName(AValue: string);
  private
    // Notify event handlers for view addition
    procedure NotifyViewAdded(AView: TDocView);
    // Notify event handlers for view removal
    procedure NotifyViewRemoved(AView: TDocView);
    // Notify event handlers that the last view was removed
    procedure NotifyLastViewRemoved(var FreeDocument: Boolean);
    // Notify event handlers for modification state changes
    procedure NotifyModifiedStateChange;
    // Notify event handlers for filename changes
    procedure NotifyFileNameChanged;
    // Notify event handlers for document saves
    procedure NotifySaved(AFileName: string);
    // Notify event handlers for document loads
    procedure NotifyLoaded(AFileName: string);
    // Called by given view when it is being destroyed
    procedure ViewDestroyed(AView: TDocView); virtual;
  protected
    // Called when the document has been registered with a document manager
    procedure Registered; virtual;
    // Called when the document has been unregistered from a document manager
    procedure Unregistered(ADocumentManager: TDocumentManager); virtual;
    // Create a new view of the given view class for this document, by default
    // this will just create a new instance of the given class
    function CreateView(AViewClass: TDocViewClass): TDocView; virtual;
    // Called when a view is added to this document
    procedure ViewAdded(AView: TDocView); virtual;
    // Called when a view is removed from this document
    procedure ViewRemoved(AView: TDocView); virtual;
    // Called when the last view is removed, if FreeDocument is True the
    // document will free itself
    procedure LastViewRemoved(var FreeDocument: Boolean); virtual;
    // Called when the document's modified state is changed
    procedure ModifiedStateChanged; virtual;
    // Called when the document's filename is changed
    procedure FileNameChanged; virtual;
    // Called after the document was saved to the given filename
    procedure Saved(AFileName: string); virtual;
    // Called after the document was loaded from the given filename
    procedure Loaded(AFileName: string); virtual;
    // Called to check if the document can be saved to the given file
    function CanSaveDocumentToFile(AFileName: string): Boolean; virtual;
    // Called to save the document to the given filename
    procedure SaveDocumentToFile(AFileName: string); virtual;
    // Called to check if the document can be loaded from the given file
    function CanLoadDocumentFromFile(AFileName: string): Boolean; virtual;
    // Called to load the document from the given filename
    procedure LoadDocumentFromFile(AFileName: string); virtual;
  public
    // Create the document
    constructor Create; virtual;
    // Destroy the document, will automatically destroy all views and unregister
    // the document from the manager
    destructor Destroy; override;
    // Returns true if the document can be modified
    function CanModify: Boolean; virtual;
    // Returns true if the document's filename can be changed to the given one
    function CanChangeFileName(ANewFileName: string): Boolean; virtual;
    // Mark the document as modified, returns True if the state changed
    function MarkModified: Boolean;
    // Remove the modified flag (mark the document as not being modified)
    procedure ClearModified;
    // Create a new view of the given class and initialize it
    function NewView(AViewClass: TDocViewClass): TDocView;
    // Destroy all the document's views
    procedure DestroyAllViews;
    // Refresh all the document's views
    procedure RefreshViews;
    // Invalidate all the document's views
    procedure InvalidateViews(Detail: TObject=nil);
    // Returns true if the document has any views
    function HasViews: Boolean; inline;
    // Returns the index of the given view or -1 if not found
    function IndexOfView(AView: TDocView): Integer; inline;
    // Save the document to the current filename (FileName must be set)
    procedure Save;
    // Load the document from the current filename (FileName must be set)
    procedure Load;
    // Check if the document can be saved to the given filename
    function CanSaveToFile(AFileName: string): Boolean;
    // Save the document to the given filename, if UpdateFileName is True then
    // the given filename becomes the new filename of the document and the
    // document's modified status is cleared
    procedure SaveToFile(AFileName: string; UpdateFileName: Boolean=True);
    // Check if the document can be loaded from the given filename
    function CanLoadFromFile(AFileName: string): Boolean;
    // Read the document from the given filename, if UpdateFileName is True then
    // the given filename becomes the new filename of the document and the
    // document's modified status is cleared
    procedure LoadFromFile(AFileName: string; UpdateFileName: Boolean=True);
  public
    // The views of this document
    property Views[AIndex: Integer]: TDocView read GetViews;
    // The number of views this document has
    property ViewCount: Integer read GetViewCount;
    // The document's filename
    property FileName: string read FFileName write SetFileName;
    // The document manager that owns this document
    property DocumentManager: TDocumentManager read FDocumentManager;
    // If True the document has been modified
    property Modified: Boolean read GetModified;
  public
    // Called when a view is added
    property OnViewAdd: TViewEvent read FOnViewAdd write FOnViewAdd;
    // Called when a view is removed
    property OnViewRemove: TViewEvent read FOnViewRemove write FOnViewRemove;
    // Called when the last view is removed, if FreeDocument is True the
    // document will free itself
    property OnLastViewRemove: TDocumentLastViewRemovedEvent read FOnLastViewRemove write FOnLastViewRemove;
    // Called when the modified state changes
    property OnModify: TNotifyEvent read FOnModify write FOnModify;
    // Called when the filename changes
    property OnFileNameChange: TNotifyEvent read FOnFileNameChange write FOnFileNameChange;
    // Called after the document is saved to the given filename
    property OnSave: TFileNameEvent read FOnSave write FOnSave;
    // Called after the document has been loaded from the given filename
    property OnLoad: TFileNameEvent read FOnLoad write FOnLoad;
  end;

  { TDocumentManager }
  TDocumentManager = class(TComponent)
  private
    // Property storage
    FDocuments: TDocumentDynArray;
    FListeners: specialize TDynArray<TDocManagerListener>;
    FOnDocumentRegister: TDocumentEvent;
    FOnDocumentUnregister: TDocumentEvent;
    FOnDocumentModify: TDocumentEvent;
    FOnDocumentFileNameChange: TDocumentEvent;
    FOnDocumentSave: TDocumentFileEvent;
    FOnDocumentLoad: TDocumentFileEvent;
    FOnViewAdd: TViewEvent;
    FOnViewRemove: TViewEvent;
    function GetDocuments(AIndex: Integer): TDocument; inline;
    function GetDocumentCount: Integer; inline;
  private
    FOnLastViewRemove: TDocManagerLastViewRemovedEvent;
    // Called by the document when the modified state is changed
    procedure DocumentModifiedStateChanged(ADocument: TDocument);
    // Called by the document when the filename is changed
    procedure DocumentFileNameChanged(ADocument: TDocument);
    // Called by the document after it was saved to the given filename
    procedure DocumentSaved(ADocument: TDocument; AFileName: string);
    // Called by the document after it was loaded from the given filename
    procedure DocumentLoaded(ADocument: TDocument; AFileName: string);
    // Called by the document when a view is added
    procedure ViewAdded(AView: TDocView);
    // Called by the document when a view is removed
    procedure ViewRemoved(AView: TDocView);
    // Called by the document when the last view is removed
    procedure LastViewRemoved(ADocument: TDocument; var FreeDocument: Boolean);
  public
    // Create the manager
    constructor Create(AOwner: TComponent); override;
    // Destroy the manager, will automatically destroy any registered documents
    destructor Destroy; override;
    // Register the given document with the manager, unless unregistered the
    // document will be automatically destroyed when the manager is destroyed
    procedure RegisterDocument(ADocument: TDocument);
    // Unregister the given document from the manager
    procedure UnregisterDocument(ADocument: TDocument);
    // Destroy all registered documents in the manager
    procedure DestroyAllDocuments;
    // Destroy all views (can cause some or all documents to be destroyed too)
    procedure DestroyAllViews;
    // Returns true if there are any documents in the manager
    function HasDocuments: Boolean; inline;
    // Return the index of the given document or -1 if not found
    function IndexOfDocument(ADocument: TDocument): Integer; inline;
    // Add a document manager event listener
    procedure AddListener(AListener: TDocManagerListener);
    // Remove a document manager event listener
    procedure RemoveListener(AListener: TDocManagerListener);
  public
    // The documents registered with the manager
    property Documents[AIndex: Integer]: TDocument read GetDocuments;
    // The number of documents registered with the manager
    property DocumentCount: Integer read GetDocumentCount;
  public
    // Called when a document is registered
    property OnDocumentRegister: TDocumentEvent read FOnDocumentRegister write FOnDocumentRegister;
    // Called when a document is unregistered
    property OnDocumentUnregister: TDocumentEvent read FOnDocumentUnregister  write FOnDocumentUnregister;
    // Called when a document's modified state changes
    property OnDocumentModify: TDocumentEvent read FOnDocumentModify write FOnDocumentModify;
    // Called when a document's filename changes
    property OnDocumentFileNameChange: TDocumentEvent read FOnDocumentFileNameChange write FOnDocumentFileNameChange;
    // Called after a document is saved to the given filename
    property OnDocumentSave: TDocumentFileEvent read FOnDocumentSave write FOnDocumentSave;
    // Called after a document has been loaded from the given filename
    property OnDocumentLoad: TDocumentFileEvent read FOnDocumentLoad write FOnDocumentLoad;
    // Called when a view is added to a document
    property OnViewAdd: TViewEvent read FOnViewAdd write FOnViewAdd;
    // Called when a view is removed from a document
    property OnViewRemove: TViewEvent read FOnViewRemove write FOnViewRemove;
    // Called when the last view of a document is removed, if FreeDocument is
    // True then the document will free itself
    property OnLastViewRemove: TDocManagerLastViewRemovedEvent read FOnLastViewRemove write FOnLastViewRemove;
  end;

implementation

{ TDocManagerListener }
procedure TDocManagerListener.ViewAdded(ADocument: TDocument; AView: TDocView);
begin
end;

procedure TDocManagerListener.ViewRemoved(ADocument: TDocument; AView: TDocView);
begin
end;

procedure TDocManagerListener.LastViewRemoved(ADocument: TDocument; var FreeDocument: Boolean);
begin
end;

procedure TDocManagerListener.DocumentUnregistered(ADocument: TDocument);
begin
end;

procedure TDocManagerListener.DocumentRegistered(ADocument: TDocument);
begin
end;

procedure TDocManagerListener.DocumentModified(ADocument: TDocument);
begin
end;

procedure TDocManagerListener.DocumentFileNameChanged(ADocument: TDocument);
begin
end;

procedure TDocManagerListener.DocumentSaved(ADocument: TDocument;
  AFileName: string);
begin
end;

procedure TDocManagerListener.DocumentLoaded(ADocument: TDocument;
  AFileName: string);
begin
end;

{ TDocView }
constructor TDocView.Create;
begin
  inherited Create;
end;

destructor TDocView.Destroy;
begin
  if Assigned(FDocument) then FDocument.ViewDestroyed(Self);
  inherited Destroy;
end;

procedure TDocView.InitializeView;
begin
end;

procedure TDocView.Modified;
begin
end;

procedure TDocView.Invalidated(Detail: TObject);
begin
  // Subclasses can do a better job using Detail if needed
  Refresh;
end;

procedure TDocView.FileNameChanged;
begin
end;

procedure TDocView.Saved(AFileName: string);
begin
end;

procedure TDocView.Loaded(AFileName: string);
begin
end;

procedure TDocView.Refresh;
begin
  // Subclasses must make this *do* something
end;

procedure TDocView.Invalidate(Detail: TObject);
begin
  if Assigned(FDocument) then FDocument.InvalidateViews(Detail);
end;

procedure TDocView.RefreshViews;
begin
  if Assigned(FDocument) then FDocument.RefreshViews;
end;

function TDocView.IsLastView: Boolean;
begin
  if Assigned(Document) then
    Result:=Document.GetViewCount=1
  else
    Result:=True;
end;

{ TControlDocView }
function TControlDocView.HasControl: Boolean;
begin
  Result:=Assigned(GetControl);
end;

{ TDocument }
constructor TDocument.Create;
begin
  inherited Create;
end;

destructor TDocument.Destroy;
begin
  FFlags += [TDocFlag.Destroying];
  DestroyAllViews;
  if Assigned(FDocumentManager) then FDocumentManager.UnregisterDocument(Self);
  inherited Destroy;
end;

function TDocument.GetModified: Boolean;
begin
  Result:=TDocFlag.Modified in FFlags;
end;

function TDocument.GetViews(AIndex: Integer): TDocView;
begin
  Result:=FViews[AIndex];
end;

function TDocument.GetViewCount: Integer;
begin
  Result:=FViews.Count;
end;

procedure TDocument.SetFileName(AValue: string);
begin
  if FFileName=AValue then Exit;
  if not CanChangeFileName(AValue) then
    raise EDocumentError.Create('Cannot change the filename to ' + AValue);
  FFileName:=AValue;
  NotifyFileNameChanged;
end;

procedure TDocument.NotifyViewAdded(AView: TDocView);
begin
  ViewAdded(AView);
  if Assigned(FDocumentManager) then FDocumentManager.ViewAdded(AView);
  if Assigned(FOnViewAdd) then FOnViewAdd(Self, AView);
end;

procedure TDocument.NotifyViewRemoved(AView: TDocView);
begin
  ViewRemoved(AView);
  if Assigned(FDocumentManager) then FDocumentManager.ViewRemoved(AView);
  if Assigned(FOnViewRemove) then FOnViewRemove(Self, AView);
end;

procedure TDocument.NotifyLastViewRemoved(var FreeDocument: Boolean);
begin
  LastViewRemoved(FreeDocument);
  if Assigned(FDocumentManager) then FDocumentManager.LastViewRemoved(Self, FreeDocument);
  if Assigned(FOnLastViewRemove) then FOnLastViewRemove(Self, FreeDocument);
end;

procedure TDocument.NotifyModifiedStateChange;
var
  I: Integer;
begin
  ModifiedStateChanged;
  for I:=0 to FViews.Count - 1 do FViews[I].Modified;
  if Assigned(FDocumentManager) then FDocumentManager.DocumentModifiedStateChanged(Self);
  if Assigned(FOnModify) then FOnModify(Self);
end;

procedure TDocument.NotifyFileNameChanged;
var
  I: Integer;
begin
  FileNameChanged;
  for I:=0 to FViews.Count - 1 do FViews[I].FileNameChanged;
  if Assigned(FDocumentManager) then FDocumentManager.DocumentFileNameChanged(Self);
  if Assigned(FOnFileNameChange) then FOnFileNameChange(Self);
end;

procedure TDocument.NotifySaved(AFileName: string);
var
  I: Integer;
begin
  Saved(AFileName);
  for I:=0 to FViews.Count - 1 do FViews[I].Saved(AFileName);
  if Assigned(FDocumentManager) then FDocumentManager.DocumentSaved(Self, AFileName);
  if Assigned(FOnSave) then FOnSave(Self, AFileName);
end;

procedure TDocument.NotifyLoaded(AFileName: string);
var
  I: Integer;
begin
  Loaded(AFileName);
  for I:=0 to FViews.Count - 1 do FViews[I].Loaded(AFileName);
  if Assigned(FDocumentManager) then FDocumentManager.DocumentLoaded(Self, AFileName);
  if Assigned(FOnLoad) then FOnLoad(Self, AFileName);
end;

procedure TDocument.ViewDestroyed(AView: TDocView);
var
  FreeDocument: Boolean;
begin
  AView.FDocument:=nil;
  FViews.Remove(AView);
  NotifyViewRemoved(AView);
  if FViews.Count=0 then begin
    FreeDocument:=True;
    NotifyLastViewRemoved(FreeDocument);
    if FreeDocument and not (TDocFlag.Destroying in FFlags) then Free;
  end;
end;

procedure TDocument.Registered;
begin
end;

procedure TDocument.Unregistered(ADocumentManager: TDocumentManager);
begin
end;

function TDocument.CreateView(AViewClass: TDocViewClass): TDocView;
begin
  Result:=TDocView(AViewClass.NewInstance);
  Result.Create;
end;

procedure TDocument.ViewAdded(AView: TDocView);
begin
end;

procedure TDocument.ViewRemoved(AView: TDocView);
begin
end;

procedure TDocument.LastViewRemoved(var FreeDocument: Boolean);
begin
end;

procedure TDocument.ModifiedStateChanged;
begin
end;

procedure TDocument.FileNameChanged;
begin
end;

procedure TDocument.Saved(AFileName: string);
begin
end;

procedure TDocument.Loaded(AFileName: string);
begin
end;

function TDocument.CanSaveDocumentToFile(AFileName: string): Boolean;
begin
  Result:=True;
end;

procedure TDocument.SaveDocumentToFile(AFileName: string);
begin
  raise EDocumentError.Create('Document saving is not implemented');
end;

function TDocument.CanLoadDocumentFromFile(AFileName: string): Boolean;
begin
  Result:=True;
end;

procedure TDocument.LoadDocumentFromFile(AFileName: string);
begin
  raise EDocumentError.Create('Document loading is not implemented');
end;

function TDocument.CanModify: Boolean;
begin
  Result:=Assigned(DocumentManager);
end;

function TDocument.CanChangeFileName(ANewFileName: string): Boolean;
begin
  Result:=Assigned(DocumentManager);
end;

function TDocument.MarkModified: Boolean;
begin
  if Modified then Exit(True);
  if not CanModify then Exit(False);
  FFlags += [TDocFlag.Modified];
  NotifyModifiedStateChange;
  Result:=Modified;
end;

procedure TDocument.ClearModified;
begin
  if not Modified then Exit;
  FFlags -= [TDocFlag.Modified];
  NotifyModifiedStateChange;
end;

function TDocument.NewView(AViewClass: TDocViewClass): TDocView;
begin
  Assert(Assigned(AViewClass), 'Missing view class');
  Result:=CreateView(AViewClass);
  if Assigned(Result) then begin
    FViews.Add(Result);
    Result.FDocument:=Self;
    Result.InitializeView;
    NotifyViewAdded(Result);
  end;
end;

procedure TDocument.DestroyAllViews;
begin
  while FViews.Count > 0 do FViews.Last.Free;
end;

procedure TDocument.RefreshViews;
var
  I: Integer;
begin
  for I:=0 to FViews.Count - 1 do FViews[I].Refresh;
end;

procedure TDocument.InvalidateViews(Detail: TObject);
var
  I: Integer;
begin
  for I:=0 to FViews.Count - 1 do FViews[I].Invalidated(Detail);
end;

function TDocument.HasViews: Boolean;
begin
  Result:=FViews.Count > 0;
end;

function TDocument.IndexOfView(AView: TDocView): Integer;
begin
  Assert(Assigned(AView), 'Missing view');
  Result:=FViews.IndexOf(AView);
end;

procedure TDocument.Save;
begin
  if FileName='' then raise EDocumentError.Create('No filename in document');
  SaveToFile(FileName, True);
end;

procedure TDocument.Load;
begin
  if FileName='' then raise EDocumentError.Create('No filename in document');
  LoadFromFile(FileName, True);
end;

function TDocument.CanSaveToFile(AFileName: string): Boolean;
begin
  if AFileName='' then Exit(False);
  if FileExists(AFileName) and FileIsReadOnly(AFileName) then Exit(False);
  Result:=CanSaveDocumentToFile(AFileName);
end;

procedure TDocument.SaveToFile(AFileName: string; UpdateFileName: Boolean);
begin
  if not CanSaveToFile(AFileName) then
    raise EDocumentError.Create('Cannot save the document to "' + AFileName + '"');
  try
    SaveDocumentToFile(AFileName);
  finally
    if UpdateFileName then begin
      FileName:=AFileName;
      ClearModified;
    end;
    NotifySaved(AFileName);
  end;
end;

function TDocument.CanLoadFromFile(AFileName: string): Boolean;
begin
  if AFileName='' then Exit(False);
  if not FileExists(AFileName) then Exit(False);
  Result:=CanLoadDocumentFromFile(AFileName);
end;

procedure TDocument.LoadFromFile(AFileName: string; UpdateFileName: Boolean);
begin
  if not CanLoadFromFile(AFileName) then
    raise EDocumentError.Create('Cannot load the document from "' + AFileName + '"');
  try
    LoadDocumentFromFile(AFileName);
  finally
    if UpdateFileName then begin
      FileName:=AFileName;
      ClearModified;
    end;
    NotifyLoaded(AFileName);
  end;
end;

{ TDocumentManager }
constructor TDocumentManager.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

destructor TDocumentManager.Destroy;
begin
  DestroyAllDocuments;
  inherited Destroy;
end;

procedure TDocumentManager.DocumentModifiedStateChanged(ADocument: TDocument);
var
  I: Integer;
begin
  if Assigned(FOnDocumentModify) then FOnDocumentModify(Self, ADocument);
  for I:=0 to FListeners.Count - 1 do FListeners[I].DocumentModified(ADocument);
end;

procedure TDocumentManager.DocumentFileNameChanged(ADocument: TDocument);
var
  I: Integer;
begin
  if Assigned(FOnDocumentFileNameChange) then FOnDocumentFileNameChange(Self, ADocument);
  for I:=0 to FListeners.Count - 1 do FListeners[I].DocumentFileNameChanged(ADocument);
end;

procedure TDocumentManager.DocumentSaved(ADocument: TDocument; AFileName: string);
var
  I: Integer;
begin
  if Assigned(FOnDocumentSave) then FOnDocumentSave(Self, ADocument, AFileName);
  for I:=0 to FListeners.Count - 1 do FListeners[I].DocumentSaved(ADocument, AFileName);
end;

procedure TDocumentManager.DocumentLoaded(ADocument: TDocument; AFileName: string);
var
  I: Integer;
begin
  if Assigned(FOnDocumentLoad) then FOnDocumentLoad(Self, ADocument, AFileName);
  for I:=0 to FListeners.Count - 1 do FListeners[I].DocumentLoaded(ADocument, AFileName);
end;

procedure TDocumentManager.ViewAdded(AView: TDocView);
var
  I: Integer;
begin
  if Assigned(FOnViewAdd) then FOnViewAdd(Self, AView);
  for I:=0 to FListeners.Count - 1 do FListeners[I].ViewAdded(AView.Document, AView);
end;

procedure TDocumentManager.ViewRemoved(AView: TDocView);
var
  I: Integer;
begin
  if Assigned(FOnViewRemove) then FOnViewRemove(Self, AView);
  for I:=0 to FListeners.Count - 1 do FListeners[I].ViewRemoved(AView.Document, AView);
end;

procedure TDocumentManager.LastViewRemoved(ADocument: TDocument;
  var FreeDocument: Boolean);
var
  I: Integer;
begin
  if Assigned(FOnLastViewRemove) then FOnLastViewRemove(Self, ADocument, FreeDocument);
  for I:=0 to FListeners.Count - 1 do FListeners[I].LastViewRemoved(ADocument, FreeDocument);
end;

function TDocumentManager.GetDocuments(AIndex: Integer): TDocument;
begin
  Result:=FDocuments[AIndex];
end;

function TDocumentManager.GetDocumentCount: Integer;
begin
  Result:=FDocuments.Count;
end;

procedure TDocumentManager.RegisterDocument(ADocument: TDocument);
var
  I: Integer;
begin
  Assert(Assigned(ADocument), 'Missing document');
  Assert(not Assigned(ADocument.FDocumentManager), 'The document is already registered in a manager');
  FDocuments.Add(ADocument);
  ADocument.FDocumentManager:=Self;
  ADocument.Registered;
  for I:=0 to FListeners.Count - 1 do FListeners[I].DocumentRegistered(ADocument);
end;

procedure TDocumentManager.UnregisterDocument(ADocument: TDocument);
var
  I: Integer;
begin
  Assert(Assigned(ADocument), 'Missing document');
  Assert(ADocument.FDocumentManager=Self, 'The document is not registered with this manager');
  FDocuments.Remove(ADocument);
  ADocument.FDocumentManager:=nil;
  ADocument.Unregistered(Self);
  for I:=0 to FListeners.Count - 1 do FListeners[I].DocumentUnregistered(ADocument);
end;

procedure TDocumentManager.DestroyAllDocuments;
begin
  while FDocuments.Count > 0 do FDocuments.Last.Free;
end;

procedure TDocumentManager.DestroyAllViews;
var
  Views: TDocViewDynArray;
  I, J: Integer;
begin
  // Collect all views
  for I:=0 to FDocuments.Count - 1 do
    for J:=0 to FDocuments[I].FViews.Count - 1 do
      Views.Add(FDocuments[I].FViews[J]);
  // Destroy the views
  for I:=0 to Views.Count - 1 do Views[I].Free;
end;

function TDocumentManager.HasDocuments: Boolean;
begin
  Result:=FDocuments.Count > 0;
end;

function TDocumentManager.IndexOfDocument(ADocument: TDocument): Integer;
begin
  Assert(Assigned(ADocument), 'Missing document');
  Result:=FDocuments.IndexOf(ADocument);
end;

procedure TDocumentManager.AddListener(AListener: TDocManagerListener);
begin
  Assert(Assigned(AListener), 'Missing listener');
  FListeners.Add(AListener);
end;

procedure TDocumentManager.RemoveListener(AListener: TDocManagerListener);
begin
  Assert(Assigned(AListener), 'Missing listener');
  FListeners.Remove(AListener);
end;

end.

