unit VertexFragmentPairs;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LResources, Forms, Controls, Graphics, Dialogs, GL, GLu,
  GLext, Maths;

type

  { TVertexFragmentPair }

  TVertexFragmentPair = class(TComponent)
  private
    FProgramID: GLuint;
    FVertexCode: TStrings;
    FFragmentCode: TStrings;
    FHasError: Boolean;
    FErrorMessage: string;
    procedure CreateProgram;
    procedure DestroyProgram;
    procedure SetVertexCode(AValue: TStrings);
    procedure SetFragmentCode(AValue: TStrings);
    function GetProgramID: GLuint;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Use: Boolean;
    procedure SetFloatUniform(AName: string; AValue: Single);
    procedure SetIntUniform(AName: string; AValue: GLint);
    procedure SetVectorUniform(AName: string; const AValue: TVector);
    procedure SetVectorUniform(AName: string; const AValue: TVector; W: Single);
    procedure SetMatrixUniform(AName: string; const AValue: TMatrix);
  published
    property VertexCode: TStrings read FVertexCode write SetVertexCode nodefault;
    property FragmentCode: TStrings read FFragmentCode write SetFragmentCode nodefault;
  public
    property ProgramID: GLuint read GetProgramID;
    property HasError: Boolean read FHasError;
    property ErrorMessage: string read FErrorMessage;
  end;

procedure Register;

implementation

type

  { TVertexFragmentPairStringList }

  TVertexFragmentPairStringList = class(TStringList)
  private
    Pair: TVertexFragmentPair;
  public
    procedure Changed; override;
  end;

{ TVertexFragmentShaderPairStringList }

procedure TVertexFragmentPairStringList.Changed;
begin
  Pair.DestroyProgram;
end;

{ TVertexFragmentPair }

procedure TVertexFragmentPair.CreateProgram;
var
  VS, FS: GLuint;
  A, B: GLint;
  CodePtr: PGLchar;
  S: AnsiString;
begin
  if csDesigning in ComponentState then Exit;
  if not Assigned(glCreateShader) then begin
    if not Load_GL_VERSION_2_1() then begin
      FHasError:=True;
      FErrorMessage:='Cannot load OpenGL 2.1';
    end;
    Exit;
  end;
  VS:=glCreateShader(GL_VERTEX_SHADER);
  FS:=glCreateShader(GL_FRAGMENT_SHADER);
  S:=FVertexCode.Text;
  A:=Length(S);
  CodePtr:=PGLchar(S);
  glShaderSource(VS, 1, @CodePtr, @A);
  S:=FFragmentCode.Text;
  A:=Length(S);
  CodePtr:=PGLchar(S);
  glShaderSource(FS, 1, @CodePtr, @A);
  glGetError();
  glCompileShader(VS);
  if glGetError() <> GL_NO_ERROR then begin
    glDeleteShader(VS);
    glDeleteShader(FS);
    glGetError();
    FHasError:=True;
    FErrorMessage:='Error compiling vertex shader';
    Exit;
  end;
  glCompileShader(FS);
  if glGetError() <> GL_NO_ERROR then begin
    glDeleteShader(VS);
    glDeleteShader(FS);
    glGetError();
    FHasError:=True;
    FErrorMessage:='Error compiling fragment shader';
    Exit;
  end;
  glGetShaderiv(VS, GL_COMPILE_STATUS, @A);
  if A=GL_FALSE then begin
    glGetShaderiv(VS, GL_INFO_LOG_LENGTH, @A);
    SetLength(S, A);
    glGetShaderInfoLog(VS, A, @B, @S[1]);
    glDeleteShader(VS);
    glDeleteShader(FS);
    glGetError();
    FHasError:=True;
    FErrorMessage:='Error compiling vertex shader: ' + S;
    Exit;
  end;
  glGetShaderiv(FS, GL_COMPILE_STATUS, @A);
  if A=GL_FALSE then begin
    glGetShaderiv(FS, GL_INFO_LOG_LENGTH, @A);
    SetLength(S, A);
    glGetShaderInfoLog(FS, A, @B, @S[1]);
    glDeleteShader(VS);
    glDeleteShader(FS);
    glGetError();
    FHasError:=True;
    FErrorMessage:='Error compiling fragment shader: ' + S;
    Exit;
  end;
  DestroyProgram;
  glGetError();
  FProgramID:=glCreateProgram();
  glAttachShader(FProgramID, VS);
  glAttachShader(FProgramID, FS);
  glLinkProgram(FProgramID);
  glDeleteShader(VS);
  glDeleteShader(FS);
  if glGetError() <> GL_NO_ERROR then begin
    FHasError:=True;
    FProgramID:=0;
    FErrorMessage:='Error linking GLSL program';
    Exit;
  end;
  glGetProgramiv(FProgramID, GL_LINK_STATUS, @A);
  if A=GL_FALSE then begin
    glGetProgramiv(FProgramID, GL_INFO_LOG_LENGTH, @A);
    SetLength(S, A);
    glGetProgramInfoLog(FProgramID, A, @B, @S[1]);
    FProgramID:=0;
    glGetError();
    FHasError:=True;
    FErrorMessage:='Error compiling fragment shader: ' + S;
    Exit;
  end;
  FHasError:=False;
end;

procedure TVertexFragmentPair.DestroyProgram;
begin
  if FProgramID=0 then Exit;
  glDeleteProgram(FProgramID);
  FProgramID:=0;
end;

function TVertexFragmentPair.GetProgramID: GLuint;
begin
  if FProgramID=0 then CreateProgram;
  Result:=FProgramID;
end;

constructor TVertexFragmentPair.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FVertexCode:=TVertexFragmentPairStringList.Create;
  TVertexFragmentPairStringList(FVertexCode).Pair:=Self;
  FFragmentCode:=TVertexFragmentPairStringList.Create;
  TVertexFragmentPairStringList(FFragmentCode).Pair:=Self;
  FVertexCode.Add('#version 120');
  FVertexCode.Add('');
  FVertexCode.Add('void main()');
  FVertexCode.Add('{');
  FVertexCode.Add('    gl_Position = ftransform();');
  FVertexCode.Add('}');
  FFragmentCode.Add('#version 120');
  FFragmentCode.Add('');
  FFragmentCode.Add('void main()');
  FFragmentCode.Add('{');
  FFragmentCode.Add('    gl_FragColor = vec4(1.0, 1.0, 1.0, 1.0);');
  FFragmentCode.Add('}');
end;

procedure TVertexFragmentPair.SetFragmentCode(AValue: TStrings);
begin
  if FFragmentCode.Text=AValue.Text then Exit;
  FFragmentCode.Assign(AValue);
  DestroyProgram;
end;

procedure TVertexFragmentPair.SetVertexCode(AValue: TStrings);
begin
  if FVertexCode.Text=AValue.Text then Exit;
  FVertexCode.Assign(AValue);
  DestroyProgram;
end;

destructor TVertexFragmentPair.Destroy;
begin
  DestroyProgram;
  FreeAndNil(FFragmentCode);
  FreeAndNil(FVertexCode);
  inherited Destroy;
end;

function TVertexFragmentPair.Use: Boolean;
begin
  if FProgramID=0 then CreateProgram;
  if FProgramID <> 0 then glUseProgram(FProgramID);
  Result:=not HasError;
end;

procedure TVertexFragmentPair.SetFloatUniform(AName: string;
  AValue: Single);
begin
  glUniform1f(glGetUniformLocation(ProgramID, PGLchar(AName)), AValue);
end;

procedure TVertexFragmentPair.SetIntUniform(AName: string; AValue: GLint);
begin
  glUniform1i(glGetUniformLocation(ProgramID, PGLchar(AName)), AValue);
end;

procedure TVertexFragmentPair.SetVectorUniform(AName: string;
  const AValue: TVector);
begin
  glUniform3f(glGetUniformLocation(ProgramID, PGLchar(AName)), AValue.x, AValue.y, AValue.z);
end;

procedure TVertexFragmentPair.SetVectorUniform(AName: string;
  const AValue: TVector; W: Single);
begin
  glUniform4f(glGetUniformLocation(ProgramID, PGLchar(AName)), AValue.x, AValue.y, AValue.z, W);
end;

procedure TVertexFragmentPair.SetMatrixUniform(AName: string;
  const AValue: TMatrix);
var
  Arr: T16dArray;
begin
  Arr:=AValue.ToArray;
  glUniformMatrix4fv(glGetUniformLocation(ProgramID, PGLchar(AName)), 1, GL_FALSE, @Arr);
end;

procedure Register;
begin
  {$I vertexfragmentpairs_icon.lrs}
  RegisterComponents('RTTK',[TVertexFragmentPair]);
end;

end.
